/*
 * Copyright 2016 - 2025 Gary Talent (gary@drinkingtea.net). All rights reserved.
 */

#include "rotatecommand.hpp"

namespace nostalgia::gfx {

static void rotateLeft(TileSheet::SubSheet &ss, ox::Point const &pt, int const depth = 0) noexcept {
	if (depth >= 4) {
		return;
	}
	auto const h = ss.rows * TileHeight;
	auto const dstPt = ox::Point{pt.y, h - 1 - pt.x};
	auto const srcIdx = ptToIdx(pt, ss.columns);
	auto const dstIdx = ptToIdx(dstPt, ss.columns);
	auto const src = ss.pixels[srcIdx];
	auto &dst = ss.pixels[dstIdx];
	rotateLeft(ss, dstPt, depth + 1);
	dst = src;
}

static void rotateLeft(TileSheet::SubSheet &ss) noexcept {
	auto const w = ss.columns * TileWidth;
	auto const h = ss.rows * TileHeight;
	for (int x = 0; x < w / 2; ++x) {
		for (int y = 0; y < h / 2; ++y) {
			rotateLeft(ss, {w - 1 - y, x});
		}
	}
}

static void rotateRight(TileSheet::SubSheet &ss, ox::Point const &pt, int const depth = 0) noexcept {
	if (depth >= 4) {
		return;
	}
	auto const w = ss.columns * TileWidth;
	auto const dstPt = ox::Point{w - 1 - pt.y, pt.x};
	auto const srcIdx = ptToIdx(pt, ss.columns);
	auto const dstIdx = ptToIdx(dstPt, ss.columns);
	auto const src = ss.pixels[srcIdx];
	auto &dst = ss.pixels[dstIdx];
	rotateRight(ss, dstPt, depth + 1);
	dst = src;
}

static void rotateRight(TileSheet::SubSheet &ss) noexcept {
	auto const w = ss.columns * TileWidth;
	auto const h = ss.rows * TileHeight;
	for (int x = 0; x < w / 2; ++x) {
		for (int y = 0; y < h / 2; ++y) {
			rotateRight(ss, {w - 1 - y, x});
		}
	}
}


RotateCommand::RotateCommand(
	TileSheet &img,
	TileSheet::SubSheetIdx idx,
	Direction const dir) noexcept:
		m_img(img),
		m_idx(std::move(idx)),
		m_dir{dir} {
}

ox::Error RotateCommand::redo() noexcept {
	switch (m_dir) {
		case Direction::Left:
			rotateLeft(getSubSheet(m_img, m_idx));
			break;
		case Direction::Right:
			rotateRight(getSubSheet(m_img, m_idx));
			break;
	}
	return {};
}

ox::Error RotateCommand::undo() noexcept {
	switch (m_dir) {
		case Direction::Left:
			rotateRight(getSubSheet(m_img, m_idx));
			break;
		case Direction::Right:
			rotateLeft(getSubSheet(m_img, m_idx));
			break;
	}
	return {};
}

int RotateCommand::commandId() const noexcept {
	return static_cast<int>(CommandId::Rotate);
}

TileSheet::SubSheetIdx const&RotateCommand::subsheetIdx() const noexcept {
	return m_idx;
}

}
